import asyncio

from aiohttp import web, http_websocket
from nacl.exceptions import BadSignatureError
from nacl.signing import VerifyKey

from v6d0auth import certs, cdb

__all__ = ('get_app',)


def define_routes(routes: web.RouteTableDef, loop: asyncio.AbstractEventLoop):
    print(certs.vkey.encode().hex())
    mycdb = cdb.CDB(loop)
    loop.create_task(mycdb.job())

    @routes.get('/')
    async def home(_request: web.Request):
        return web.Response(body='v6d0auth\n')

    @routes.post('/approve')
    async def approve(request: web.Request):
        try:
            cert = mycdb.approve(await request.read())
        except BadSignatureError:
            raise web.HTTPUnauthorized
        except KeyError:
            raise web.HTTPNotFound
        else:
            return web.Response(body=cert)

    @routes.post('/push')
    async def push(request: web.Request):
        try:
            timeout = mycdb.push(VerifyKey(await request.read())).timeout
        except KeyError:
            raise web.HTTPTooManyRequests
        else:
            return web.Response(body=str(timeout))

    @routes.post('/pull')
    async def pull(request: web.Request):
        try:
            cert = mycdb.pull(VerifyKey(await request.read()))
        except KeyError:
            raise web.HTTPNotFound
        else:
            return web.Response(body=cert)

    @routes.get('/pullws')
    async def pullws(request: web.Request):
        ws = web.WebSocketResponse()
        await ws.prepare(request)
        try:
            srq = mycdb.requester_mapping[VerifyKey(await ws.receive_bytes())]
        except TypeError:
            await ws.close(code=http_websocket.WSCloseCode.POLICY_VIOLATION)
        else:
            try:
                cert = await srq.future
            except asyncio.CancelledError:
                if not srq.future.cancelled():
                    srq.future.cancel()
                if not srq.cancelled:
                    srq.future = asyncio.get_event_loop().create_future()
                await ws.close(code=http_websocket.WSCloseCode.TRY_AGAIN_LATER)
            else:
                await ws.send_bytes(cert)
                await ws.close()
        return ws


def app_routes(loop: asyncio.AbstractEventLoop) -> web.RouteTableDef:
    routes = web.RouteTableDef()
    define_routes(routes, loop)
    return routes


def app_with_routes(routes: web.RouteTableDef):
    app = web.Application()
    app.add_routes(routes)
    return app


def get_app(loop: asyncio.AbstractEventLoop) -> web.Application:
    return app_with_routes(app_routes(loop))
