import asyncio
import functools
import heapq
import time
import weakref
from typing import MutableMapping, Optional

from nacl.signing import VerifyKey
from nacl.utils import random

from v6d0auth import certs

__all__ = ('CDB',)

TIMEOUT = 300


@functools.total_ordering
class SignatureRequest:
    def __init__(self, requester: VerifyKey, loop: asyncio.AbstractEventLoop):
        self._requester = requester
        self.timeout = time.time() + TIMEOUT
        self.handle: bytes = random(12)
        self.approved: Optional[bytes] = None
        self.cancelled = False
        self.future: asyncio.Future[bytes] = loop.create_future()

    def __le__(self, other):
        if isinstance(other, SignatureRequest):
            return self.timeout < other.timeout
        else:
            return NotImplemented

    def timed_out(self) -> bool:
        return time.time() > self.timeout

    def approve(self) -> bytes:
        if self.approved is None:
            self.approved = certs.sign(bytes(self._requester))
            self.future.set_result(self.approved)
            print('approved', self.handle.hex())
        return self.approved

    def cancel(self):
        if not self.future.done():
            self.future.cancel()
            self.cancelled = True


class CDB:
    def __init__(self, loop: asyncio.AbstractEventLoop):
        self.handle_mapping: MutableMapping[bytes, SignatureRequest] = weakref.WeakValueDictionary()
        self.requester_mapping: MutableMapping[VerifyKey, SignatureRequest] = weakref.WeakValueDictionary()
        self.heap: list[SignatureRequest] = []
        self._loop = loop

    def _cleanup(self):
        while self.heap and self.heap[0].timed_out():
            request = heapq.heappop(self.heap)
            request.cancel()
            yield request

    def cleanup(self):
        if self.heap:
            print('cleaning up')
        for request in self._cleanup():
            print('cleaned', request.handle.hex())

    def push(self, requester: VerifyKey) -> SignatureRequest:
        if requester in self.requester_mapping:
            raise KeyError
        request = SignatureRequest(requester, self._loop)
        self.requester_mapping[requester] = request
        heapq.heappush(self.heap, request)
        self.handle_mapping[request.handle] = request
        print('requested', request.handle.hex(), requester.encode().hex())
        return request

    def _approve(self, handle: bytes) -> bytes:
        return self.handle_mapping[handle].approve()

    def approve(self, data: bytes) -> bytes:
        handle = certs.verify(data)
        return self._approve(handle)

    def pull(self, vkey: VerifyKey) -> Optional[bytes]:
        return self.requester_mapping[vkey].approved

    async def job(self):
        while True:
            await asyncio.sleep(TIMEOUT)
            self.cleanup()
